package cd.frontend.semantic;

import cd.frontend.semantic.SemanticAnalyzer.SemanticLocation;
import cd.frontend.semantic.SemanticFailure.Cause;
import cd.ir.Ast;
import cd.ir.Ast.*;
import cd.ir.ExprVisitor;
import cd.ir.Symbol;
import cd.ir.Symbol.*;

import java.util.List;

/**
 * Semantic analyzer for {@link Expr} nodes in the pattern of a visitor. <br>
 * All methods may throw {@link SemanticFailure} if an error is discovered,
 * with the appropriate {@link Cause}. <br>
 * The current implementation returns the resulting {@link Symbol} for each expression and
 * uses the {@link ClassDecl} argument to obtain the self-reference in the {@link ThisRef} node.
 * <br> Some methods may throw a {@link RuntimeException} if they are in a
 * illegal situation, given the grammar of Javali and current implementation.
 * This signals an error in the implementation, not a user-caused error.
 * @see cd.frontend.semantic.StmtAnalyzer
 * @author Carlos Galindo
 * @author Levin Moser
 */
public class ExprAnalyzer extends ExprVisitor<TypeSymbol,SemanticLocation> {
	/** Reference to the main semantic analyzer, used to access the type list via
	 * {@link SemanticAnalyzer#findType(String)} */
	private SemanticAnalyzer sem;

	ExprAnalyzer(SemanticAnalyzer sem) {
		this.sem = sem;
	}

	/**
	 * Checks the possible semantic errors that can appear in a {@link BinaryOp} node
	 * @param ast The node describing the binary operation
	 * @param arg The class and method in which the operation is performed
	 * @return A {@link PrimitiveTypeSymbol} of resulting type, depending on the operation
	 * performed (either {@code int} or {@code boolean})
	 * @throws SemanticFailure If the types of both operands aren't equal
	 * @throws SemanticFailure With {@code ==} or {@code !=} operators, if no argument's type is
	 * a subtype of the other
	 * @throws SemanticFailure If the type of the operands doesn't match the type
	 * of the operation (either {@code int} or {@code boolean}
	 * @throws RuntimeException If the implementation has missed any {@link BinaryOp.BOp} type
	 */
	@Override
	public TypeSymbol binaryOp(BinaryOp ast, SemanticLocation arg) {
		TypeSymbol leftType = visit(ast.left(), arg);
		TypeSymbol rightType = visit(ast.right(), arg);
		TypeSymbol desiredType;
		TypeSymbol returnType = PrimitiveTypeSymbol.booleanType;
		switch (ast.operator) {
			case B_AND:
			case B_OR:
				desiredType = PrimitiveTypeSymbol.booleanType;
				break;
			case B_PLUS:
			case B_MINUS:
			case B_TIMES:
			case B_DIV:
			case B_MOD:
				desiredType = PrimitiveTypeSymbol.intType;
				returnType = desiredType;
				break;
			case B_LESS_THAN:
			case B_GREATER_THAN:
			case B_LESS_OR_EQUAL:
			case B_GREATER_OR_EQUAL:
				desiredType = PrimitiveTypeSymbol.intType;
				break;
			case B_EQUAL:
			case B_NOT_EQUAL:
				// Can take any pair of types that are a subtype of each other
				desiredType = null;
				break;
			default:
				throw new RuntimeException("Missing BOp in implementation of ExprAnalyzer#binaryOp(...)");
		}

		if (desiredType == null) {
			if (leftType.isSubtypeOf(rightType) || rightType.isSubtypeOf(leftType))
				return returnType;
			else
				throw new SemanticFailure(Cause.TYPE_ERROR,
						"Operation \"%s\" operation with incompatible types %s and %s",
						ast.operator.repr, leftType, rightType);
		} else if (leftType == rightType) {
			if (leftType == desiredType)
				return returnType;
			else
				throw new SemanticFailure(Cause.TYPE_ERROR,
						"Arguments of %s operation must be of %s type, found %s",
						ast.operator.repr, desiredType, leftType);
		} else {
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Arguments of %s operation must be of same type %s, found %s and %s",
					ast.operator.repr, desiredType, leftType, rightType);
		}
	}

	/**
	 * Obtains the {@code boolean} type
	 * @param ast Node describing the constant
	 * @param arg Class and method in which the expression exists
	 * @return {@link PrimitiveTypeSymbol} representing the {@code boolean} type
	 */
	@Override
	public TypeSymbol booleanConst(BooleanConst ast, SemanticLocation arg) {
		return PrimitiveTypeSymbol.booleanType;
	}

	/**
	 * Obtains the {@code int} type (all reads in Javali result in an integer)
	 * @param ast Node describing the expression
	 * @param arg Class and method in which the expression exists
	 * @return {@link PrimitiveTypeSymbol} representing the {@code int} type
	 */
	@Override
	public TypeSymbol builtInRead(BuiltInRead ast, SemanticLocation arg) {
		return PrimitiveTypeSymbol.intType;
	}

	/**
	 * Checks possible semantic errors in a {@link Cast} node. <br>
	 * Uses {@link TypeSymbol#isSubtypeOf(TypeSymbol)} to determine if the types are compatible.
	 * @return The resulting type after the cast
	 * @throws SemanticFailure If the destination type doesn't exist
	 * @throws SemanticFailure If the types are incompatible (none is a subtype of the other)
	 */
	@Override
	public TypeSymbol cast(Cast ast, SemanticLocation arg) {
		TypeSymbol castType = sem.findType(ast.typeName);
		TypeSymbol exprType = visit(ast.arg(), arg);
		if (exprType.isSubtypeOf(castType) || castType.isSubtypeOf(exprType))
			return castType;
		else
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Type %s cannot be casted to %s, as they are unrelated", exprType, castType);
	}

	/**
	 * Checks possible semantic errors in a {@link Field} node.
	 * @param ast Node describing the field access
	 * @param arg Class and method in which the expression exists
	 * @return Field accessed, as defined in the corresponding {@link ClassSymbol}
	 * @throws SemanticFailure If the field to access doesn't exist
	 * @throws SemanticFailure If the expression in whose field is accessed
	 * is not a {@link ClassSymbol}
	 */
	@Override
	public TypeSymbol field(Field ast, SemanticLocation arg) {
		TypeSymbol variableType = visit(ast.arg(), arg);
		if (!(variableType instanceof ClassSymbol))
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"The field of a %s variable cannot be accessed, it is not a object type", variableType);
		ClassSymbol classType = (ClassSymbol) variableType;
		VariableSymbol field = classType.getField(ast.fieldName);
		if (field == null)
			throw new SemanticFailure(Cause.NO_SUCH_FIELD,
					"The field %s doesn't exist in class %s",
					ast.fieldName, variableType);
		else
			return field.type;
	}

	/**
	 * Checks possible semantic errors in a {@link Index} node.
	 * @param ast Node describing the index operation
	 * @param arg Class and method in which the expression exists
	 * @return Type of the elements of the array indexed
	 * @throws SemanticFailure If the index is not an integer
	 * @throws SemanticFailure If the expression indexed is not an array
	 */
	@Override
	public TypeSymbol index(Index ast, SemanticLocation arg) {
		TypeSymbol arrayType = visit(ast.left(), arg);
		TypeSymbol indexType = visit(ast.right(), arg);

		if (indexType != PrimitiveTypeSymbol.intType)
			throw new SemanticFailure(Cause.TYPE_ERROR, "The index must be a int, found %s", indexType);
		else if (!(arrayType instanceof ArrayTypeSymbol))
			throw new SemanticFailure(Cause.TYPE_ERROR, "Expecting array type, found %s", arrayType);
		else
			return ((ArrayTypeSymbol) arrayType).elementType;
	}

	/**
	 * Obtains the {@code int} type
	 * @param ast Node describing the constant
	 * @param arg Class and method in which the expression exists
	 * @return A PrimitiveTypeSymbol for the type int
	 */
	@Override
	public TypeSymbol intConst(IntConst ast, SemanticLocation arg) {
		return PrimitiveTypeSymbol.intType;
	}

	/**
	 * Checks that the method referenced exists in the current scope and the typing of
	 * the parameters.
	 * @param ast Node describing the method call
	 * @param arg Class and method in which the expression exists
	 * @return The return type for the method called
	 * @throws SemanticFailure If the number of parameters doesn't match the declaration
	 * @throws SemanticFailure If the parameters' types don't match the declaration
	 * @throws SemanticFailure If the method doesn't exist
	 * @throws SemanticFailure If the {@link MethodCallExpr#receiver()} is not
	 * a object type (methods cannot be called on PrimitiveTypes or Arrays)
	 */
	@Override
	public TypeSymbol methodCall(MethodCallExpr ast, SemanticLocation arg) {
		TypeSymbol receiverType = visit(ast.receiver(), arg);
		// Check receiver validity
		if (!(receiverType instanceof ClassSymbol))
			throw new SemanticFailure(Cause.TYPE_ERROR, "Expected object class, found %s", receiverType);
		// Find method symbol
		ast.sym = sem.findMethod(ast.methodName, (ClassSymbol) receiverType);
		if (ast.sym == null)
			throw new SemanticFailure(Cause.NO_SUCH_METHOD, "Method %s(...) doesn't exist", ast.methodName);
		// Compare expected with available parameters
		List<Expr> realParams = ast.argumentsWithoutReceiver();
		List<VariableSymbol> expectedParams = ast.sym.parameters;
		if (realParams.size() != expectedParams.size())
			throw new SemanticFailure(Cause.WRONG_NUMBER_OF_ARGUMENTS,
					"%s call has %d parameters, expected %d",
					ast.sym, realParams.size(), expectedParams.size());
		for (int i = 0; i < realParams.size(); i++) {
			TypeSymbol realType = visit(realParams.get(i), arg);
			TypeSymbol expectedType = expectedParams.get(i).type;
			if (!realType.isSubtypeOf(expectedType))
				throw new SemanticFailure(Cause.TYPE_ERROR, "Parameter %d of %s of type %s, expected %s",
						i, ast.sym, realType, expectedType);
		}
		return ast.sym.returnType;
	}

	/**
	 * Checks for semantic errors in the creation of a new object of the specified type
	 * @param ast Node describing the new object expression
	 * @param arg Class and method in which the expression exists
	 * @return {@link ClassSymbol} corresponding to the type of the new object
	 * @throws SemanticFailure If the type referenced doesn't exist
	 */
	@Override
	public TypeSymbol newObject(NewObject ast, SemanticLocation arg) {
		return sem.findType(ast.typeName);
	}

	/**
	 * Checks for semantic errors in the creation of a new array of the specified type and size
	 * @param ast Node describing the new array expression
	 * @param arg Class and method in which the expression exists
	 * @return ArrayTypeSymbol of the corresponding type
	 * @throws SemanticFailure If the array type doesn't exist
	 * @throws SemanticFailure If the expression indicating the size of the array is not an {@code int}
	 */
	@Override
	public TypeSymbol newArray(NewArray ast, SemanticLocation arg) {
		// Type exists (if it does, it is guaranteed by the grammar to be a ArrayTypeSymbol)
		ArrayTypeSymbol type = (ArrayTypeSymbol) sem.findType(ast.typeName);
		// Array length must be of type int
		TypeSymbol indexType = visit(ast.arg(), arg);
		if (indexType != PrimitiveTypeSymbol.intType)
			throw new SemanticFailure(Cause.TYPE_ERROR, "Array size must be of type int, found %s", indexType);
		else
			return type;
	}

	/**
	 * Obtains the {@code null} type
	 * @param ast Node describing the new object instantiation expression
	 * @param arg Class and method in which the expression exists
	 * @return The ClassSymbol corresponding to {@code null}
	 */
	@Override
	public TypeSymbol nullConst(NullConst ast, SemanticLocation arg) {
		return ClassSymbol.nullType;
	}

	/**
	 * Obtains the reference to {@code this}
	 * @param ast The AST node
	 * @param arg Class and method in which the class reference is accessed
	 * @return The VariableSymbol {@code this} from the class
	 */
	@Override
	public TypeSymbol thisRef(ThisRef ast, SemanticLocation arg) {
		return arg.classSymbol.thisSymbol.type;
	}

	/**
	 * Checks whether a {@link UnaryOp} is valid semantically
	 * @param ast The node describing the operation
	 * @param arg Class and method in which the expression exists
	 * @return The resulting type of the operation
	 * @throws SemanticFailure With "{@code !}" operator, if the operand is not a {@code boolean}
	 * @throws SemanticFailure With "{@code +}" or "{@code -}" operators, if the operand is not an {@code int}
	 */
	@Override
	public TypeSymbol unaryOp(UnaryOp ast, SemanticLocation arg) {
		TypeSymbol argType = visit(ast.arg(), arg);

		TypeSymbol desiredType = PrimitiveTypeSymbol.intType;
		if (ast.operator.equals(Ast.UnaryOp.UOp.U_BOOL_NOT))
			desiredType = PrimitiveTypeSymbol.booleanType;

		if (argType == desiredType)
			return desiredType;
		else
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Operator %s expected operand of type %s, found %s",
					ast.operator.repr, desiredType, argType);
	}

	/**
	 * Checks whether the variable referenced by the {@link Var} node exists
	 * in the current scope, either as local variable, method parameter or field
	 * (including inherited fields)
	 * <br> In the process, the {@link Var} node is linked to the
	 * {@link VariableSymbol}
	 * @param ast Node describing the variable access
	 * @param arg Class and method that define the current scope in which the
	 *            variable is looked for
	 * @return The {@link VariableSymbol} corresponding to the variable
	 * @throws SemanticFailure If the variable doesn't exist in the current scope
	 */
	@Override
	public TypeSymbol var(Var ast, SemanticLocation arg) {
		// Local variable
		ast.sym = arg.getMethod().locals.get(ast.name);
		if (ast.sym != null)
			return ast.sym.type;
		// Method parameter
		for(VariableSymbol var : arg.getMethod().parameters)
			if (var.name.equals(ast.name))
				return (ast.sym = var).type;
		// Class field (or field of a superclass)
		ast.sym = arg.classSymbol.getField(ast.name);
		if (ast.sym != null)
			return ast.sym.type;
		// The variable cannot be found in any known scope
		throw new SemanticFailure(Cause.NO_SUCH_VARIABLE, "Variable %s not found", ast.name);
	}
}
