package cd.frontend.semantic;

import cd.Main;
import cd.frontend.semantic.SemanticFailure.Cause;
import cd.ir.Ast;
import cd.ir.Ast.ClassDecl;
import cd.ir.Ast.MethodDecl;
import cd.ir.Symbol;
import cd.ir.Symbol.*;

import java.util.*;

public class SemanticAnalyzer {
	/** Name of the main class of the program, containing the start point for its execution */
	public static final String MAIN_CLASS_NAME = "Main";
	/** Name of the start point method inside the main class, with signature {@code void MAIN_METHOD_NAME()} */
	public static final String MAIN_METHOD_NAME = "main";

	/** Reference to main method, for main class and list of types */
	public final Main main;
	/** Statement analyzer, entrypoint for the method body analysis */
	public final StmtAnalyzer sa = new StmtAnalyzer(this);
	/** For analyzing expressions, used by the StmtAnalyzer when an expression is found */
	public final ExprAnalyzer ea = new ExprAnalyzer(this);

	private List<ClassDecl> classDecls;

	public SemanticAnalyzer(Main main) {
		this.main = main;
	}

	/**
	 * Entrypoint for semantic analysis, will check that a list of classes is correct semantically
	 * @param classDeclList List of classes that form the Javali program
	 * @throws SemanticFailure If any semantic error is found
	 */
	public void check(List<ClassDecl> classDeclList) throws SemanticFailure {
		this.classDecls = classDeclList;
		checkCircularInheritance(classDecls);
		// Built-in type generation
		main.allTypeSymbols = new ArrayList<>();
		addBuiltInTypes(main.allTypeSymbols);
		// Main semantic analysis
		for (ClassDecl classDecl : classDecls)
			sa.visit(classDecl, null);
		// Find entrypoint Main and Main.main(...)
		findMainClass();
		findEntrypoint(main.mainType);
	}

	/**
	 * Generates the basic types that exist in any Javali program by default
	 * @param list Type list to which the types must be added
	 */
	private void addBuiltInTypes(List<TypeSymbol> list) {
		// Add primitive types
		list.add(PrimitiveTypeSymbol.booleanType);
		list.add(PrimitiveTypeSymbol.intType);
		list.add(PrimitiveTypeSymbol.voidType);
		// Add basic class types
		list.add(ClassSymbol.objectType);
		list.add(ClassSymbol.nullType);
		// Add arrays of primitive and class types
		list.add(new ArrayTypeSymbol(PrimitiveTypeSymbol.intType));
		list.add(new ArrayTypeSymbol(PrimitiveTypeSymbol.booleanType));
		list.add(new ArrayTypeSymbol(ClassSymbol.objectType));
	}

	/**
	 * Finds the {@code Main} class in a Javali program
	 * @throws SemanticFailure If there is no such class
	 */
	private void findMainClass() {
		for (TypeSymbol type : main.allTypeSymbols)
			if (type.name.equals(MAIN_CLASS_NAME))
				main.mainType = (ClassSymbol) type;
		if (main.mainType == null)
			throw new SemanticFailure(Cause.INVALID_START_POINT,
					"Can't find class called \"%s\"", MAIN_CLASS_NAME);
	}

	/**
	 * Finds the entrypoint for the Javali program
	 * @param mainType {@code Main} class of the program, where the entrypoint should be.
	 * @throws SemanticFailure If there is no such entrypoint, of it has incorrect return
	 * value or parameters.
	 */
	private void findEntrypoint(ClassSymbol mainType) {
		MethodSymbol mainMethod = mainType.getMethod(MAIN_METHOD_NAME);
		if (mainMethod == null)
			throw new SemanticFailure(Cause.INVALID_START_POINT,
					"No method called \"%s\" in %s class", MAIN_METHOD_NAME, MAIN_CLASS_NAME);
		else if (mainMethod.returnType != PrimitiveTypeSymbol.voidType)
			throw new SemanticFailure(Cause.INVALID_START_POINT,
					"Method %s.%s(...) returns %s, expected void",
					MAIN_CLASS_NAME, MAIN_METHOD_NAME, mainMethod.returnType);
		else if (!mainMethod.parameters.isEmpty())
			throw new SemanticFailure(Cause.INVALID_START_POINT,
					"Method %s.%s(...) must have no arguments, found %d",
					MAIN_CLASS_NAME, MAIN_METHOD_NAME, mainMethod.parameters.size());
	}

	/**
	 * Looks up the requested type in the list of types. If not found, will try to read the remaining
	 * {@link ClassDecl} nodes to find the desired type.
	 * @param type Name of the type. Can be built-in, user-declared, or an array type.
	 * @return The TypeSymbol object associated with the type (no new symbol is created, only looked up from
	 * the general list).
	 * @throws SemanticFailure If the type doesn't exist
	 */
	public TypeSymbol findType(String type) {
		// Option 1: type already exists
		for (TypeSymbol symbol : main.allTypeSymbols)
			if (type.equals(symbol.name))
				return symbol;
		// Option 2: type is declared by user but hasn't been visited yet
		String noArrayType = type;
		if (type.contains("[]")) // To find the type of an array, remove "[]" and search for the name
			noArrayType = type.substring(0, type.indexOf("[]"));
		for (ClassDecl classDecl : classDecls)
			if (classDecl.name.equals(noArrayType)) {
				sa.visit(classDecl, null);
				return findType(type); // Will use option 1
			}
		// Option 3: the type doesn't exist
		throw new SemanticFailure(Cause.NO_SUCH_TYPE, "Type %s doesn't exist", type);
	}

	/**
	 * Looks up the requested method in the corresponding class. If not found, will try to find it as
	 * {@link MethodDecl} inside the {@link ClassDecl}. Will also look in superclasses of {@code receiverType}.
	 * @param methodName Name of the method that is being looked up
	 * @param receiverType Class in which the method must be found
	 * @return The method symbol (not created, but looked up from the class' method list.
	 */
	public MethodSymbol findMethod(String methodName, ClassSymbol receiverType) {
		// Option 0: looking for methods in Object or null class (that has no methods or associated ast)
		if (receiverType == ClassSymbol.objectType || receiverType == ClassSymbol.nullType)
			throw new SemanticFailure(Cause.NO_SUCH_METHOD, "Type %s has no methods", receiverType);
		// Option 1: method has already been discovered
		MethodSymbol res = receiverType.methods.get(methodName);
		if (res != null) return res;
		// Option 2: method exists but hasn't been read yet (look in superclasses)
		assert receiverType.ast != null;
		for (MethodDecl methodDecl : receiverType.ast.methods()) {
			if (methodDecl.name.equals(methodName)) {
				sa.visit(methodDecl, new SemanticLocation(receiverType));
				return findMethod(methodName, receiverType); // Will use Option 1
			}
		}
		// Option 3: method exists in superclass
		res = findMethod(methodName, receiverType.superClass);
		if (res != null) return res;
		// Option 4: method doesn't exist
		throw new SemanticFailure(Cause.NO_SUCH_METHOD, "Method %s of class %s doesn't exist",
				methodName, receiverType.name);
	}

	/**
	 * Recursively checks for circular inheritance <br>
	 * This check is done with the Strings extracted from the list of class declarations, as to perform it
	 * before the type check (which can be difficult to deal with given undiscovered circular inheritance), in order
	 * to avoid methods such as {@link ClassSymbol#getMethod(String)} or {@link ClassSymbol#getField(String)}, which
	 * are recursive by nature, throw a {@link StackOverflowError}.
	 * @param classDecls List of class declarations to check against
	 * @throws SemanticFailure If circular inheritance is found
	 * @throws SemanticFailure If any type inherited by a class doesn't exist (guaranteed to be a class type by
	 * the language grammar).
	 */
	public void checkCircularInheritance(List<ClassDecl> classDecls) {
		Map<String, String> inherits = new HashMap<>();
		for (ClassDecl classDecl : classDecls)
			inherits.put(classDecl.name, classDecl.superClass);
		for (ClassDecl classDecl : classDecls)
			checkCircularInheritance(classDecl.name, new HashSet<>(), inherits);
	}

	private void checkCircularInheritance(String className, Set<String> set, Map<String, String> inherits) {
		if (className.equals("Object")) return;
		if (!set.add(className))
			throw new SemanticFailure(Cause.CIRCULAR_INHERITANCE,
					"Class %s has itself as superclass", className);
		if (inherits.get(className) != null)
			checkCircularInheritance(inherits.get(className), set, inherits);
		else
			throw new SemanticFailure(Cause.NO_SUCH_TYPE, "Class %s doesn't exist", className);
	}

	/**
	 * Helper class that stores the position of a statement or expression being visited, by providing a link
	 * to the method and class symbols. This is useful when visiting variable nodes, method calls
	 * and {@code this} references. <br>
	 * When entering and existing the body of a method, {@link SemanticLocation#beginMethod(MethodSymbol)} and
	 * {@link SemanticLocation#endMethod()} are used to set and unset the methodSymbol field.
	 */
	public static class SemanticLocation {
		private MethodSymbol methodSymbol;
		public final ClassSymbol classSymbol;

		/**
		 * Creates a semantic location that points to the {@code classSymbol}
		 */
		SemanticLocation(ClassSymbol classSymbol) {
			this(classSymbol, null);
		}

		private SemanticLocation(ClassSymbol classSymbol, MethodSymbol methodSymbol) {
			assert classSymbol != null;
			this.methodSymbol = methodSymbol;
			this.classSymbol = classSymbol;
		}

		public MethodSymbol getMethod() {
			return methodSymbol;
		}

		public boolean isInMethod() {
			return methodSymbol != null;
		}

		/**
		 * Setter for methodSymbol
		 */
		public void beginMethod(MethodSymbol methodSymbol) {
			this.methodSymbol = methodSymbol;
		}

		/**
		 * Unsetter for methodSymbol
		 */
		public void endMethod() {
			this.methodSymbol = null;
		}
	}
}
