package cd.frontend.semantic;

import cd.frontend.semantic.SemanticAnalyzer.SemanticLocation;
import cd.frontend.semantic.SemanticFailure.Cause;
import cd.ir.Ast;
import cd.ir.Ast.*;
import cd.ir.AstVisitor;
import cd.ir.Symbol;
import cd.ir.Symbol.*;
import cd.ir.Symbol.VariableSymbol.Kind;

/**
 * Semantic analyzer for {@link Ast} nodes in the pattern of a visitor. <br>
 * All methods may throw {@link SemanticFailure} if an error is discovered,
 * with the appropriate {@link Cause} and a short explanation to give some context
 * to the user. <br>
 * The current implementation employs the {@link Symbol} returned to signal a
 * statement, or group of them that contain a unconditional return (or return in both
 * branches of a {@link IfElse} node. The type is checked by
 * {@link StmtAnalyzer#returnStmt(ReturnStmt, SemanticLocation)} and the existence, by
 * {@link StmtAnalyzer#methodDecl(MethodDecl, SemanticLocation)} <br>
 * Uses the {@link SemanticLocation} argument to give context to each statement and
 * expression visited, so that they may access variables, fields and methods.
 * @see ExprAnalyzer
 * @author Carlos Galindo
 * @author Levin Moser
 */
public class StmtAnalyzer extends AstVisitor<Symbol,SemanticLocation> {
	/**
	 * Reference to the main semantic analyzer <br>
	 * Needed to access {@link SemanticAnalyzer#findType(String)} and the
	 * {@link ExprAnalyzer}
	 */
	private SemanticAnalyzer sa;

	StmtAnalyzer(SemanticAnalyzer sa) {
		this.sa = sa;
	}

	/**
	 * Checks semantic errors in a {@link Assign} node
	 * @param ast The node containing all the information
	 * @param arg The class and method in which the statement takes place
	 * @return {@code null} (No return can occur in an assigment statement)
	 * @throws SemanticFailure If the right-hand side is not a subtype of the left-hand side
	 * @throws SemanticFailure If the left-hand side is a method call or a {@code this} reference
	 */
	@Override
	public Symbol assign(Assign ast, SemanticLocation arg) {
		TypeSymbol right = sa.ea.visit(ast.right(), arg);
		TypeSymbol left  = sa.ea.visit(ast.left(), arg);
		if (!right.isSubtypeOf(left))
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"In assignment, %s is not a subtype of %s", right, left);
		// Only variables, fields and array dereferences are valid as left-hand side
		if (ast.left() instanceof Var)
			return null;
		if (ast.left() instanceof Field)
			return null;
		if (ast.left() instanceof Index)
			return null;
		String leftExprType;
		if (ast.left() instanceof MethodCallExpr)
			leftExprType = "method call";
		else if (ast.left() instanceof ThisRef)
			leftExprType = "this reference";
		else
			throw new RuntimeException("Missing valid or invalid left-hand side of assignment, check grammar");
		throw new SemanticFailure(Cause.NOT_ASSIGNABLE,
				"Left-hand side of assignment must be field, var or index, found %s", leftExprType);
	}

	/**
	 * Checks a possible error in a {@link BuiltInWrite} AST node.
	 * @param ast The node containing all the information
	 * @param arg The class and method in which the statement takes place
	 * @return {@code null} (No return statement can happen inside a write)
	 * @throws SemanticFailure If the type of the expression to write is not integer
	 */
	@Override
	public Symbol builtInWrite(BuiltInWrite ast, SemanticLocation arg) {
		TypeSymbol type = sa.ea.visit(ast.arg(), arg);
		if (type != PrimitiveTypeSymbol.intType)
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Argument of write function must be an int, found %s", type);
		else
			return null;
	}

	/**
	 * Checks a class declaration and all contained fields and methods for semantic errors. <br>
	 * Will add itself and its corresponding array type to the types list.
	 * @param ast The node describing the class declaration
	 * @param arg {@code null}, as nested classes aren't permitted by the Javali grammar
	 * @return {@code null} (no need to check return types)
	 * @throws SemanticFailure If the type of the class already exists
	 * @throws SemanticFailure If the name of the class is Object (base class)
	 * @throws SemanticFailure If the superclass is not defined in the Javali program
	 */
	@Override
	public Symbol classDecl(ClassDecl ast, SemanticLocation arg) {
		if (ast.sym != null) return null; // Class already visited

		// Generate first the symbol and link it to the ast node
		ast.sym = new ClassSymbol(ast);

		// No class can be named Object
		if (ast.name.equals(ClassSymbol.objectType.name))
			throw new SemanticFailure(Cause.OBJECT_CLASS_DEFINED, "The %s class cannot be redefined", ast.name);

		// All classes have unique names
		for (TypeSymbol type : sa.main.allTypeSymbols)
			if (type.name.equals(ast.sym.name))
				throw new SemanticFailure(Cause.DOUBLE_DECLARATION, "Another class exists with the name %s", type);

		// Add basic and array type to typeList (this is the only method that generates
		// a ClassSymbol or ArrayTypeSymbol)
		sa.main.allTypeSymbols.add(ast.sym);
		sa.main.allTypeSymbols.add(new ArrayTypeSymbol(ast.sym));

		// The superclass exists (will visit other classes to look it up)
		ast.sym.superClass = (ClassSymbol) sa.findType(ast.superClass);

		// Visit children (fields and methods)
		SemanticLocation thisClass = new SemanticLocation(ast.sym);
		for (VarDecl varDecl : ast.fields())
			ast.sym.fields.put(varDecl.name, (VariableSymbol) visit(varDecl, thisClass));
		for (MethodDecl methodDecl : ast.methods())
			visit(methodDecl, thisClass);

		return ast.sym;
	}

	/**
	 * Checks for semantic correctness of a {@link MethodDecl} node, regarding both the method's
	 * signature and its body. <br>
	 * For the declaration of local variables of the method, the underlying
	 * {@link Seq} node is not visited, as to keep the logic in
	 * {@link StmtAnalyzer#seq(Seq, SemanticLocation)} simpler, and only related to
	 * the accumulation of the return paths.
	 * @param ast The node describing the method declaration
	 * @param arg The class the method is being declared in
	 * @return {@code null} (no need to check return types)
	 * @throws SemanticFailure If multiple local variables share the same name
	 * @throws SemanticFailure If there is a return type but no return statement in all paths
	 * @throws SemanticFailure If the method is overloaded
	 * @throws SemanticFailure If the method overrides another incorrectly
	 * (mismatched parameter number, type or return type)
	 * @throws SemanticFailure If the return type doesn't exist
	 */
	@Override
	public Symbol methodDecl(MethodDecl ast, SemanticLocation arg) {
		if (ast.sym != null) return null; // Already visited
		ast.sym = new MethodSymbol(ast);
		arg.beginMethod(ast.sym); // Modify the location to include the method symbol

		// Check overloading (none is permitted)
		if (arg.classSymbol.methods.get(ast.name) != null)
			throw new SemanticFailure(Cause.DOUBLE_DECLARATION,
					"Another method exists with the name %s in class %s, overloading is not permitted",
					ast.name, arg);

		// Check return type
		ast.sym.returnType = sa.findType(ast.returnType);

		// Check params type and generate VariableSymbols
		assert ast.argumentNames.size() == ast.argumentTypes.size(); // This should be guaranteed by parser
		for (int i = 0; i < ast.argumentNames.size(); i++) {
			// check for repeated param name
			for (VariableSymbol var : ast.sym.parameters)
				if (var.name.equals(ast.argumentNames.get(i)))
					throw new SemanticFailure(Cause.DOUBLE_DECLARATION,
							"There are two parameters named %s in method %s", var, ast.sym);
			// add to sym param list
			ast.sym.parameters.add(new VariableSymbol(
					ast.argumentNames.get(i),
					sa.findType(ast.argumentTypes.get(i)),
					Kind.PARAM));
		}

		// Check override (only possible if parameters and return type match)
		MethodSymbol overridden = arg.classSymbol.getMethod(ast.name);
		if (overridden != null) {
			// Return value must be the same
			if (overridden.returnType != ast.sym.returnType)
				throw new SemanticFailure(Cause.INVALID_OVERRIDE,
						"Overridden methods must share return type, expected %s, found %s",
						overridden.returnType, ast.sym.returnType);
			// Number of parameters must be the same
			if (overridden.parameters.size() != ast.sym.parameters.size())
				throw new SemanticFailure(Cause.INVALID_OVERRIDE,
						"Overridden methods must have the same number of parameters, expected %d, found %d",
						overridden.parameters.size(), ast.sym.parameters.size());
			// Types must match for all parameters
			for (int i = 0; i < ast.sym.parameters.size(); i++)
				if (overridden.parameters.get(i).type != ast.sym.parameters.get(i).type)
					throw new SemanticFailure(Cause.INVALID_OVERRIDE,
							"Overridden methods must match parameters type, expected %s, found %s for parameter %d",
							overridden.parameters.get(i).type, ast.sym.parameters.get(i).type, i);
		}

		// Check all local variables
		for (Ast decl : ast.decls().children()) {
			VarDecl varDecl = (VarDecl) decl;
			VariableSymbol varSymbol = (VariableSymbol) visit(varDecl, arg);
			ast.sym.locals.put(varDecl.name, varSymbol);
		}
		// Check statements and return type
		Symbol returnSymbol = visit(ast.body(), arg);
		if (returnSymbol == null && ast.sym.returnType != PrimitiveTypeSymbol.voidType) {
			throw new SemanticFailure(Cause.MISSING_RETURN,
					"Method %s missing return statement", ast.sym);
		}

		// Add method to class Symbol node
		arg.classSymbol.methods.put(ast.name, ast.sym);

		arg.endMethod();
		return ast.sym;
	}

	/**
	 * Checks for semantic errors in a {@link VarDecl} node. It can either receive a local variable
	 * or a class field.
	 * @param ast Node containing the type and name for the variable
	 * @param arg Class [and method] in which the variable is being declared
	 * @return {@link VariableSymbol} representing the variable
	 * @throws SemanticFailure If the variable has already been declared in this class
	 * (hiding a field from a superclass is allowed)
	 * @throws SemanticFailure If the type of the variable doesn't exist
	 * @throws SemanticFailure If the local variable has already been declared
	 */
	@Override
	public Symbol varDecl(VarDecl ast, SemanticLocation arg) {
		if (arg.isInMethod()) { // Checks for local variable
			if (arg.getMethod().locals.get(ast.name) != null)
				throw new SemanticFailure(Cause.DOUBLE_DECLARATION,
						"Variable %s is already declared as a variable", ast.name);
			for (VariableSymbol var : arg.getMethod().parameters)
				if (var.name.equals(ast.name))
					throw new SemanticFailure(Cause.DOUBLE_DECLARATION,
							"Variable %s is already declared as a parameter", ast.name);
		} else { // Checks for field variable
			if (arg.classSymbol.fields.get(ast.name) != null)
				throw new SemanticFailure(Cause.DOUBLE_DECLARATION,
						"Class %s cannot have more than one variable with the same name (%s)", arg, ast.name);
		}
		ast.sym = new VariableSymbol(
				ast.name,
				sa.findType(ast.type), // Type check included here
				arg.isInMethod() ? Kind.FIELD : Kind.LOCAL);
		return ast.sym;
	}

	/**
	 * Checks for semantic errors in a {@link IfElse} node
	 * @param ast Contains all the information regarding the condition and if/else bodies
	 * @param arg The class and method in whose body the if statement exists
	 * @return The return value if both branches have one, or {@code null} otherwise
	 * @throws SemanticFailure If the condition is not of boolean type
	 */
	@Override
	public Symbol ifElse(IfElse ast, SemanticLocation arg) {
		TypeSymbol condType = sa.ea.visit(ast.condition(), arg);
		if (condType != PrimitiveTypeSymbol.booleanType)
			throw new SemanticFailure(Cause.TYPE_ERROR, "Condition of if must be int, found %s", condType);
		Symbol then      = visit(ast.then(), arg);
		Symbol otherwise = visit(ast.otherwise(), arg);

		if (then != null && otherwise != null)
			return then;
		else
			return null;
	}

	/**
	 * Checks for semantic errors in a {@link ReturnStmt} node
	 * @param ast Contains all information about the return statement
	 * @param arg The class and method in which this statement takes place
	 * @return Type of the underlying expression, to be returned by the method
	 * @throws SemanticFailure If the return type is not a subtype of the method signature's return type
	 * @throws SemanticFailure If the method does return a non-void but the return statement has no
	 * expression to return
	 */
	@Override
	public Symbol returnStmt(ReturnStmt ast, SemanticLocation arg) {
		if (ast.arg() != null) {
			TypeSymbol returnType = sa.ea.visit(ast.arg(), arg);
			if (returnType.isSubtypeOf(arg.getMethod().returnType))
				return returnType;
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Return type of %s must be a subtype of %s, found %s",
					arg.getMethod(), arg.getMethod().returnType, returnType);
		} else {
			if (arg.getMethod().returnType != PrimitiveTypeSymbol.voidType)
				throw new SemanticFailure(Cause.TYPE_ERROR,
						"Empty return, expecting type %s or subtype", arg.getMethod().returnType);
			return null;
		}
	}

	/**
	 * Checks for semantic errors in a method call
	 * @see ExprAnalyzer#methodCall(MethodCallExpr, SemanticLocation)
	 * @return {@code null} (No return statement in this node)
	 */
	@Override
	public Symbol methodCall(MethodCall ast, SemanticLocation arg) {
		sa.ea.visit(ast.getMethodCallExpr(), arg);
		return null;
	}

	/**
	 * Checks that all members of the {@link Seq} node are valid. <br>
	 * This method will only visit sequences of statements, as the
	 * sequences of variables are checked in
	 * {@link StmtAnalyzer#methodDecl(MethodDecl, SemanticLocation)}, therefore
	 * the logic is simpler and separated for unrelated elements of the method
	 * body (all local variable declaration must come before the statements)
	 * @param ast The node with the sequence of statements
	 * @param arg Class and method in which this sequence exists, to be passed
	 *            down to visited statements
	 * @return The first type returned by one of the underlying statements
	 */
	@Override
	public Symbol seq(Seq ast, SemanticLocation arg) {
		Symbol firstReturn = null;
		for (Ast aChildren : ast.children()) {
			Symbol sym = visit(aChildren, arg);
			if (sym != null && firstReturn == null)
				firstReturn = sym;
		}
		return firstReturn;
	}

	/**
	 * Checks semantic errors in a while loop
	 * @param ast Node containing all the information about the {@code while} loop
	 * @param arg The class and method in which the loop exists
	 * @return {@code null} (condition of the loop is not check to determine if
	 * it is always accessed, therefore it cannot be treated as an unconditional return)
	 * @throws SemanticFailure If the condition for the loop is not a boolean type
	 */
	@Override
	public Symbol whileLoop(WhileLoop ast, SemanticLocation arg) {
		TypeSymbol condType = sa.ea.visit(ast.condition(), arg);
		if (condType != PrimitiveTypeSymbol.booleanType)
			throw new SemanticFailure(Cause.TYPE_ERROR,
					"Condition of while loop must be int, found %s", condType);
		visit(ast.body(), arg);
		return null;
	}
}
